// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut für Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file MatrixStreams.h */

#ifndef MATRIXSTREAMS_H
#define MATRIXSTREAMS_H

#include "DOFMatrix.h"
#include "SolverMatrix.h"
#include "SystemVector.h"
#include "Collection.h"

using namespace mtl::operations;
namespace AMDiS {
  template< typename VecT, typename CurMap>
  struct VecMap {
    VecT& vec;
    CurMap& mapper;
    VecMap(VecT& vec, CurMap& mapper):
      vec(vec),mapper(mapper) {}
  };

  template< typename MatT, typename CurMap>
  struct MatMap {
    MatT& mat;
    CurMap& mapper;
    MatMap(MatT& mat, CurMap& m):
      mat(mat), mapper(m) {}
  };

  template< typename MTLMatrix, typename Mapper >
  void operator<<(MTLMatrix& matrix, MatMap<const SolverMatrix<Matrix<DOFMatrix* > >, Mapper >& Asolver)
  {
    const Matrix< DOFMatrix* >& A = *(Asolver.mat.getOriginalMat());
    int ns = A.getSize();

    Mapper& mapper(Asolver.mapper);
    set_to_zero(matrix);

    int nnz = 0;
    for (int rb = 0; rb < ns; ++rb)
      for (int cb = 0; cb < ns; ++cb)
        if (A[rb][cb])
	  nnz += A[rb][cb]->getBaseMatrix().nnz();	  
      
    {
      typedef mtl::matrix::mapped_inserter< typename Collection< MTLMatrix >::Inserter, Mapper > Inserter;
      Inserter ins(matrix, mapper, int(1.2 * nnz / matrix.dim1()));
      for (int rb = 0; rb < ns; ++rb) {
        mapper.setRow(rb);
        for (int cb = 0; cb < ns; ++cb) {
	  mapper.setCol(cb);
          if (A[rb][cb])
	    ins << A[rb][cb]->getBaseMatrix();
        }
      }
    }
  }

  template< typename Vector, typename CurMap >
  void operator<<(Vector& dest, VecMap<DOFVector<double >, CurMap >& rhs)
  {
    DOFVector<double>::Iterator it_x(&rhs.vec, USED_DOFS);
    int counter(0);
    typedef typename Collection< Vector >::Inserter Inserter ;

    Inserter swapIns(dest);
    mtl::vector::mapped_inserter< Inserter, CurMap > ins(swapIns, rhs.mapper);

    for (it_x.reset(); !it_x.end(); ++it_x) {
      ins[counter] << *it_x;
      counter++;
    }
  }

  template< typename Vector, typename CurMap >
  inline void operator>>(const Vector& dest, VecMap<DOFVector< double >, CurMap >& rhs)
  {
    DOFVector<double>::Iterator it_x(&rhs.vec, USED_DOFS);
    int counter(0);
    {
      mtl::vector::extracter< Vector > extracter(dest);
      for (it_x.reset(); !it_x.end(); ++it_x) {
	extracter[rhs.mapper.row(counter)] >> *it_x ;
	counter++;
      }
    }
  }

  template< typename Vector, typename CurMap >
  void operator<<(Vector& dest, VecMap<SystemVector, CurMap>& rhs)
  {
    int ns = rhs.vec.getSize();  // Number of systems.

      // Copy vectors
    typedef typename Collection< Vector >::Inserter Inserter;
    Inserter swapInserter(dest);
    mtl::vector::mapped_inserter< Inserter, CurMap > ins(swapInserter, rhs.mapper);
    for (int i = 0; i < ns; i++) {
      DOFVector<double>::Iterator it_source(rhs.vec.getDOFVector(i), USED_DOFS);
      int counter(0);
      rhs.mapper.setRow(i);
      for (it_source.reset(); !it_source.end(); ++it_source) {
        ins[counter] << *it_source;
        counter++;
      }
    }
  }

  template< typename Vector, typename CurMap >
  void operator>>(const Vector& dest, VecMap<SystemVector, CurMap>& rhs) 
  {
    int ns = rhs.vec.getSize();  // Number of systems.

      // Copy vectors
    for (int i = 0; i < ns; i++) {
      DOFVector< double >& dofvec(*(rhs.vec.getDOFVector(i)));
      rhs.mapper.setRow(i);
      VecMap< DOFVector< double >, CurMap > swap(dofvec, rhs.mapper);
      dest >> swap;
    }
  }
}
#endif
