#include "CahnHilliard_RB.h"
#include "Views.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"
#include "Tools.h"

using namespace AMDiS;

CahnHilliard_RB::CahnHilliard_RB(const std::string &name_) :
  super(name_),
  useMobility(false),
  doubleWell(0),
  gamma(1.0),
  eps(0.1),
  minusEps(-0.1),
  epsInv(10.0),
  minusEpsInv(-10.0),
  epsSqr(0.01),
  minusEpsSqr(-0.01),
  reinit(NULL)
{
  // parameters for CH
  Parameters::get(name + "->use mobility", useMobility); // mobility
  Parameters::get(name + "->gamma", gamma); // mobility
  Parameters::get(name + "->epsilon", eps); // interface width
  
  // type of double well: 0= [0,1], 1= [-1,1]
  Parameters::get(name + "->double-well type", doubleWell); 

  // transformation of the parameters
  minusEps = -eps;
  epsInv = 1.0/eps;
  minusEpsInv = -epsInv;
  epsSqr = sqr(eps);
  minusEpsSqr = -epsSqr;
}


CahnHilliard_RB::~CahnHilliard_RB() 
{ FUNCNAME("CahnHilliard_RB::~CahnHilliard_RB()");

  if (reinit)
    delete reinit;
}


void CahnHilliard_RB::initData()
{ FUNCNAME("CahnHilliard_RB::initData()");

  // create instance redistancing class
  reinit = new HL_SignedDistTraverse("reinit", getMesh()->getDim());
  
  super::initData();
}


void CahnHilliard_RB::solveInitialProblem(AdaptInfo *adaptInfo) 
{ FUNCNAME("CahnHilliard_RB::solveInitialProblem()");

  Flag initFlag = initDataFromFile(adaptInfo);

  if (!initFlag.isSet(DATA_ADOPTED)) {
    int initialInterface = 0;
    Initfile::get(name + "->initial interface", initialInterface);
    double initialEps = eps;
    Initfile::get(name + "->initial epsilon", initialEps);

    if (initialInterface == -1) {
      /// random values
      prob->getSolution()->getDOFVector(0)->interpol(new tools::Random(0.5 - doubleWell*0.5, 1.0 + doubleWell));
    }
    else if (initialInterface == 0) {
      /// horizontale Linie
      double a= 0.0, dir= -1.0;
      Initfile::get(name + "->line->pos", a);
      Initfile::get(name + "->line->direction", dir);
      prob->getSolution()->getDOFVector(0)->interpol(new Plane(a, dir));
    }
    else if (initialInterface == 1) {
      /// schraege Linie
      double theta = m_pi/4.0;
      prob->getSolution()->getDOFVector(0)->interpol(new PlaneRotation(0.0, theta, 1.0));
      transformDOFInterpolation(prob->getSolution()->getDOFVector(0),new PlaneRotation(0.0, -theta, -1.0), new AMDiS::Min<double>);
    }
    else if (initialInterface == 2) {
      /// Ellipse
      double a= 1.0, b= 1.0;
      Initfile::get(name + "->ellipse->a", a);
      Initfile::get(name + "->ellipse->b", b);
      prob->getSolution()->getDOFVector(0)->interpol(new Ellipse(a,b));
    }
    else if (initialInterface == 3) {
      /// zwei horizontale Linien
      double a= 0.75, b= 0.375;
      Initfile::get(name + "->lines->pos1", a);
      Initfile::get(name + "->lines->pos2", b);
      prob->getSolution()->getDOFVector(0)->interpol(new Plane(a, -1.0));
      transformDOFInterpolation(prob->getSolution()->getDOFVector(0),new Plane(b, 1.0), new AMDiS::Max<double>);
    }
    else if (initialInterface == 4) {
      /// Kreis
      double radius= 1.0;
      Initfile::get(name + "->kreis->radius", radius);
      prob->getSolution()->getDOFVector(0)->interpol(new Circle(radius));
    } else if (initialInterface == 5) {
      /// Rechteck
      double width = 0.5;
      double height = 0.3;
      WorldVector<double> center; center.set(0.5);
      Initfile::get(name + "->rectangle->width", width);
      Initfile::get(name + "->rectangle->height", height);
      Initfile::get(name + "->rectangle->center", center);
      prob->getSolution()->getDOFVector(0)->interpol(new Rectangle(width, height, center));
    }

    /// create phase-field from signed-dist-function
    if (doubleWell == 0) {
      transformDOF(prob->getSolution()->getDOFVector(0),
        new SignedDistToPhaseField(initialEps));
    } else {
      transformDOF(prob->getSolution()->getDOFVector(0),
        new SignedDistToCh(initialEps));
    }
  }
}


void CahnHilliard_RB::fillOperators()
{ FUNCNAME("CahnHilliard_RB::fillOperators()");
  MSG("CahnHilliard_RB::fillOperators()\n");
  
  // dt(c) = laplace(mu) - u*grad(c)
  // -----------------------------------
  prob->addTimeOperator(0,0);				/// < dt(c) , psi >
  
  // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
  Operator *opChLM = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
  if (useMobility) {
    if (doubleWell == 0) {
      // jacobian operators
      opChLM->addTerm(new VecAtQP_SOT(
	prob->getUnVec(0),
	new MobilityCH0(gamma)));
      prob->addMatrixOperator(*opChLM, 0, 1);
      
      Operator *opChLM2 = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
      opChLM2->addTerm(new VecGrad_FOT(
	prob->getUnVec(0),
	prob->getUnVec(1),
	new MobilityCH0Diff(gamma)), GRD_PSI);
      prob->addMatrixOperator(*opChLM2, 0, 0);
      
      // rhs operators
      Operator *opChLMe = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
      opChLMe->addTerm(new VecAtQP_SOT(
	prob->getStageSolution(0),
	new MobilityCH0(gamma)));
      opChLMe->setUhOld(prob->getStageSolution(1));
      prob->addVectorOperator(*opChLMe, 0, &minus1, &minus1);
    } else if (doubleWell == 1) {
      // jacobian operators
      opChLM->addTerm(new VecAtQP_SOT(
	prob->getUnVec(0),
	new MobilityCH1(gamma)));
      prob->addMatrixOperator(*opChLM, 0, 1);
      
      Operator *opChLM2 = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
      opChLM2->addTerm(new VecGrad_FOT(
	prob->getUnVec(0),
	prob->getUnVec(1),
	new MobilityCH1Diff(gamma)), GRD_PSI);
      prob->addMatrixOperator(*opChLM2, 0, 0);
      
      // rhs operators
      Operator *opChLMe = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
      opChLMe->addTerm(new VecAtQP_SOT(
	prob->getStageSolution(0),
	new MobilityCH1(gamma)));
      opChLMe->setUhOld(prob->getStageSolution(1));
      prob->addVectorOperator(*opChLMe, 0, &minus1, &minus1);
    }
  } else {
    opChLM->addTerm(new Simple_SOT(gamma));
    opChLM->setUhOld(prob->getStageSolution(1));
    prob->addVectorOperator(*opChLM, 0, &minus1, &minus1);
    prob->addMatrixOperator(*opChLM, 0, 1);                /// < phi*grad(mu) , grad(psi) >
  }

  // mu + eps^2*laplace(c) - M'(c)
  // ----------------------------------------------------------------------
  /// < phi*mu , psi >
  Operator *opChMu = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChMu->addZeroOrderTerm(new Simple_ZOT);
  opChMu->setUhOld(prob->getStageSolution(1));
  prob->addVectorOperator(*opChMu, 1);
  prob->addMatrixOperator(*opChMu, 1, 1, &minus1, &minus1); 
  
  /// < -eps^2*phi*grad(c) , grad(psi) >
  Operator *opChLC = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChLC->addSecondOrderTerm(new Simple_SOT(minusEpsSqr));
  opChLC->setUhOld(prob->getStageSolution(0));
  prob->addVectorOperator(*opChLC, 1);
  prob->addMatrixOperator(*opChLC, 1, 0, &minus1, &minus1);   
  
  /// < -M'(c) , psi >
  if (doubleWell == 0) {
    // jacobian operators
    Operator *opChMImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMImpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getUnVec(0),
      new DoubleWell0Diff(-1.0))); // < (3c^2-3c+0.5)*c' , psi >
    prob->addMatrixOperator(*opChMImpl, 1, 0, &minus1, &minus1);   
    // rhs operators
    Operator *opChMExpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getStageSolution(0),
      new DoubleWell0(-1.0))); // < (c^3-3/2*c^2+0.5) , psi >
    prob->addVectorOperator(*opChMExpl, 1);   
  }
  else if (doubleWell == 1) {
    // jacobian operators
    Operator *opChMImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMImpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getUnVec(0),
      new DoubleWell1Diff(-1.0))); // < (3c^2-1)*c' , psi >
    prob->addMatrixOperator(*opChMImpl, 1, 0, &minus1, &minus1);   
    // rhs operators
    Operator *opChMExpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getStageSolution(0),
      new DoubleWell1(-1.0))); // < (c^3-c) , psi >
    prob->addVectorOperator(*opChMExpl, 1);   
  }
}
