#include "DiffuseDomainFsi.h"

using namespace std;
using namespace AMDiS;

DiffuseDomainFsi::DiffuseDomainFsi(const std::string &name_) :
  super(name_),
  phase(NULL),
  bcPhase(NULL),
  phaseInv(NULL),
  velocity(NULL),
  displacement(NULL),
  beta(1.0),
  epsilon(1.e-1),
  alpha(3.0),
  mu(1.0),
  viscosity(1.0),
  density(1.0),
  lambda(1.0),
  rho(1.0),
  nonLinTerm(3),
  fileWriterPhase(NULL),
  fileWriterVelocity(NULL),
  fileWriterDisplacement(NULL)
{
  
  Parameters::get(name + "->epsilon", epsilon);

  // parameters for navier-stokes
  Initfile::get(name + "->viscosity", viscosity);
  Initfile::get(name + "->density", density);
  
  Parameters::get(name + "->beta", beta);
  Parameters::get(name + "->alpha", alpha);
  
  // parameters for linear elasticity
  Parameters::get(name + "->mu", mu);
  Parameters::get(name + "->lambda", lambda);
  Parameters::get(name + "->rho", rho);
  
  // type of non-linear term: 	0... u^old*grad(u_i^old), 
  //							1... u*grad(u_i^old), 
  //							2... u^old*grad(u_i)
  //							3... u*grad(u_i^old) + u^old*grad(u_i) - u^old*grad(u_i^old)
  Parameters::get(name + "->non-linear term", nonLinTerm); 
};

DiffuseDomainFsi::~DiffuseDomainFsi()
{
  if (fileWriterPhase != NULL)
    delete fileWriterPhase;
  if (fileWriterVelocity != NULL)
    delete fileWriterVelocity;
  if (fileWriterDisplacement != NULL)
    delete fileWriterDisplacement;
  
  delete phaseInv;
  delete phaseInvDisturbed;
  delete velocity;
  delete displacement;
}

void DiffuseDomainFsi::initData()
{
  super::initData();

  phaseInv = new DOFVector<double>(getFeSpace(0), "phaseInv");
  phaseInvDisturbed = new DOFVector<double>(getFeSpace(0), "phaseInvDisturbed");
  velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
  displacement = new DOFVector<WorldVector<double> >(getFeSpace(dow+1), "displacement");
  
  dbc_factor = beta/pow(epsilon, alpha);
  MSG_DBG("dbc_factor = %f\n", dbc_factor);
  pressure_factor = -(lambda+(2.0/3.0)*mu);
  MSG("pressure_factor = %f\n", pressure_factor);
}


void DiffuseDomainFsi::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("DiffuseDomainFsi::transferInitialSolution()");

  fileWriterPhase = new FileWriter(name + "->phase->output", getFeSpace()->getMesh(), phase);
  fileWriterVelocity = new FileVectorWriter(name + "->velocity->output", getFeSpace()->getMesh(), velocity);
  fileWriterDisplacement = new FileVectorWriter(name + "->displacement->output", getFeSpace()->getMesh(), displacement);

  super::transferInitialSolution(adaptInfo);

  calcVelocity();
  calcDisplacement();

  for (int i = 0; i < prob->getNumComponents(); i++)
    prob->setExactSolution(prob->getSolution()->getDOFVector(i), i);

  writeFiles(adaptInfo, false);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx = getMesh()->getChangeIndex();
}


void DiffuseDomainFsi::initTimestep(AdaptInfo *adaptInfo)
{
  super::initTimestep(adaptInfo);
  transformDOF(1.0, phase, phaseInv, new Subtract<double>);
  transformDOF(phaseInv, 1.e-7, phaseInvDisturbed, new Max<double>);
  VtkWriter::writeFile(phaseInv, "phaseInv.vtu");
}


void DiffuseDomainFsi::closeTimestep(AdaptInfo *adaptInfo)
{
  calcVelocity();
  calcDisplacement();
  
  super::closeTimestep(adaptInfo);
}


void DiffuseDomainFsi::fillOperators()
{ FUNCNAME("DiffuseDomainFsi::fillOperators()");

  WorldVector<DOFVector<double>* > vel;
  for (size_t i = 0; i < dow; i++)
    vel[i] = prob->getSolution()->getDOFVector(i);

  TEST_EXIT(phase != NULL && phaseInv != NULL)("PhaseField is not set!\n");

  for (size_t i = 0; i < dow; ++i) {
    // time-derivative
    // ===============
    
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
    opTime->addTerm(new Simple_ZOT(density));
    prob->addMatrixOperator(*opTime, i, i, getInvTau(), getInvTau());
    /// < (1/tau)*u_i^old , psi >
    Operator *opTimeOld = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
    opTimeOld->addTerm(new Phase_ZOT(prob->getSolution()->getDOFVector(i), density));
    prob->addVectorOperator(*opTimeOld, i, getInvTau(), getInvTau());

    /// < (1/tau)*(1-phase)*eta_i , psi >
    Operator *opTimeE = new Operator(getFeSpace(dow+1+i), getFeSpace(dow+1+i));
    opTimeE->addTerm(new VecAtQP_ZOT(phaseInvDisturbed));
    prob->addMatrixOperator(*opTimeE, dow+1+i, dow+1+i, getInvTau(), getInvTau());
    /// < (1/tau)*(1-phase)*eta_i^old , psi >
    Operator *opTimeEOld = new Operator(getFeSpace(i+dow));
    opTimeEOld->addTerm(new Vec2AtQP_ZOT(phaseInvDisturbed, prob->getSolution()->getDOFVector(dow+1+i)));
    prob->addVectorOperator(*opTimeEOld, dow+1+i, getInvTau(), getInvTau());
    
    /// < (1-phase)*u , psi >
    Operator *opU = new Operator(getFeSpace(dow+1+i), getFeSpace(i));
    opU->addTerm(new Phase_ZOT(phaseInvDisturbed, -1.0));
    prob->addMatrixOperator(*opU, dow+1+i, i);
    
    // nonlinear terms
    // ===============
    
    /// < u^old*grad(u_i^old) , psi >
    if (nonLinTerm == 0 || nonLinTerm == 3) {
	  Operator *opUGradU0 = new Operator(getFeSpace(i), getFeSpace(i));
	  opUGradU0->addTerm(new WorldVec_FOT(vel, (nonLinTerm == 0 ? -1.0 : 1.0)*density), GRD_PHI);
	  opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
      prob->addVectorOperator(*opUGradU0, i);
	}
    if (nonLinTerm == 1 || nonLinTerm == 3) {
      /// < u'*grad(u_i^old) , psi >
      for (size_t j = 0; j < dow; ++j) {
        Operator *opUGradU1 = new Operator(getFeSpace(i),getFeSpace(j));
        opUGradU1->addTerm(new PartialDerivative_ZOT(
          prob->getSolution()->getDOFVector(i), j, density));
        prob->addMatrixOperator(*opUGradU1, i, j);
      }
    } 
    if (nonLinTerm == 2 || nonLinTerm == 3) {
      /// < u^old*grad(u'_i) , psi >
      for(size_t j = 0; j < dow; ++j) {
        Operator *opUGradU2 = new Operator(getFeSpace(i),getFeSpace(i));
        opUGradU2->addTerm(new VecAndPartialDerivative_FOT(
          prob->getSolution()->getDOFVector(j), j, density), GRD_PHI);
        prob->addMatrixOperator(*opUGradU2, i, i);
      }
    }

    // ====================

    /// < (1-phase) * u^old*grad(eta_i^old) , psi >
	Operator *opUGradE0 = new Operator(prob->getFeSpace(dow+1+i),prob->getFeSpace(dow+1+i));
	opUGradE0->addTerm(new WorldVecPhase_FOT(phaseInvDisturbed, vel), GRD_PHI);
	opUGradE0->setUhOld(prob->getSolution()->getDOFVector(dow+1+i));
	prob->addVectorOperator(*opUGradE0, dow+1+i);

	/// < (1-phase) * u*grad(eta_i^old) , psi >
	for (size_t j = 0; j < dow; ++j) {
	  Operator *opUGradE1 = new Operator(prob->getFeSpace(dow+1+i),prob->getFeSpace(j));
	  opUGradE1->addTerm(new VecAndPartialDerivative_ZOT(
		phaseInvDisturbed,
		prob->getSolution()->getDOFVector(dow+1+i), j));
	  prob->addMatrixOperator(*opUGradE1, dow+1+i, j);
	}
	
	/// < (1-phase) * u^old*grad(eta_i) , psi >
	for(size_t j = 0; j < dow; ++j) {
	  Operator *opUGradE2 = new Operator(prob->getFeSpace(dow+1+i),prob->getFeSpace(dow+1+i));
	  opUGradE2->addTerm(new Vec2ProductPartial_FOT(
		phaseInvDisturbed,
		prob->getSolution()->getDOFVector(j), j), GRD_PHI);
	  prob->addMatrixOperator(*opUGradE2, dow+1+i, dow+1+i);
	}
    

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity)
    addLaplaceTerm(i);
  }

  /// phase*div(u) +(1 - phase)*(p - (lambda + 2/3*mu)*div(eta))
  for (size_t i = 0; i < dow; ++i) {
    /// < phase*d_i(u_i) , psi >
    Operator *opDivU = new Operator(prob->getFeSpace(dow),prob->getFeSpace(i));
    opDivU->addTerm(new VecAndPartialDerivative_FOT(phase, i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);

    /// < (1-phase)*d_i(eta_i) , psi >
    Operator *opDivEta = new Operator(prob->getFeSpace(dow),prob->getFeSpace(dow+1+i));
    opDivEta->addTerm(new VecAndPartialDerivative_FOT(phaseInv, i, pressure_factor), GRD_PHI);
    prob->addMatrixOperator(*opDivEta, dow, dow+1+i);
  }

  /// < (1-phase)*p , psi >
  Operator *opP = new Operator(prob->getFeSpace(dow),prob->getFeSpace(dow));
  opP->addTerm(new VecAtQP_ZOT(phaseInv));
  prob->addMatrixOperator(*opP, dow, dow);
  
  for (size_t i = 0; i < dow; ++i) {
    /// dirichlet boundary condition
    /// beta/eps^alpha (bcPhase)*(eta_i - 0)
    Operator *opDbcPhaseInvU = new Operator(prob->getFeSpace(dow+1+i), prob->getFeSpace(dow+1+i));
    opDbcPhaseInvU->addTerm(new Phase_ZOT(bcPhase, dbc_factor));
    prob->addMatrixOperator(*opDbcPhaseInvU, dow+1+i, dow+1+i);
  }
}


void DiffuseDomainFsi::addLaplaceTerm(int i)
{ FUNCNAME("DiffuseDomainFsi::addLaplaceTerm()");
    
  for (size_t j = 0; j < dow; ++j) {
	/// < nu*[grad(u)+grad(u)^t] , grad(psi) >
	Operator *opLaplaceUi1 = new Operator(prob->getFeSpace(i), prob->getFeSpace(j));
	opLaplaceUi1->addTerm(new MatrixIJPhase_SOT(phase, 1-i, 1-j, viscosity));
	prob->addMatrixOperator(*opLaplaceUi1, i, j);

	/// < (mu+lambda)*[grad(u)+grad(u)^t] , grad(psi) >
	Operator *opLaplaceEi1 = new Operator(prob->getFeSpace(i), prob->getFeSpace(dow+1+j));
	opLaplaceEi1->addTerm(new MatrixIJPhase_SOT(phaseInv, 1-i, 1-j, mu+lambda));
	prob->addMatrixOperator(*opLaplaceEi1, i, dow+1+j);
  }
  
  /// < nu*grad(u_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
	opLaplaceUi->addTerm(new Phase_SOT(phase, viscosity));
  prob->addMatrixOperator(*opLaplaceUi, i, i);

  /// < mu*grad(eta_i) , grad(psi) >
  Operator *opLaplaceEi = new Operator(prob->getFeSpace(i), prob->getFeSpace(dow+1+i));
	opLaplaceEi->addTerm(new Phase_SOT(phaseInv, mu));
  prob->addMatrixOperator(*opLaplaceEi, i, dow+1+i);

  /// < p , d_i(psi) >
  Operator *opGradP = new Operator(prob->getFeSpace(i),prob->getFeSpace(dow));
  opGradP->addTerm(new VecAndPartialDerivative_FOT(phase,i, -1.0), GRD_PSI);
  prob->addMatrixOperator(*opGradP, i, dow);
  Operator *opGradP2 = new Operator(prob->getFeSpace(i),prob->getFeSpace(dow));
  opGradP2->addTerm(new PartialDerivative_ZOT(phase,i, -1.0));
  prob->addMatrixOperator(*opGradP2, i, dow);
}
