#include "NavierStokes_TaylorHood.h"

using namespace std;
using namespace AMDiS;

NavierStokes_TaylorHood::NavierStokes_TaylorHood(const std::string &name_) :
  super(name_),
  laplaceType(0),
  nonLinTerm(2),
  oldTimestep(0.0),
  viscosity(1.0),
  density(1.0),
  theta(0.5+1.e-2),
  theta1(1-theta),
  minusTheta1(-theta1),
  velocity(NULL),
  fileWriter(NULL)
{  
  // parameters for navier-stokes
  Initfile::get(name + "->viscosity", viscosity);
  Initfile::get(name + "->density", density);
  // type of laplace operator: 0... div(nu*grad(u)), 1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 
  // type of non-linear term: 0... u^old*grad(u_i^old), 1... u'*grad(u_i^old), 2... u^old*grad(u'_i)
  Initfile::get(name + "->non-linear term", nonLinTerm); 

  Initfile::get(name + "->theta", theta); 
  theta1 = 1.0-theta;
  minusTheta1 = -theta1;

  force.set(0.0);
  Initfile::get(name + "->force", force);

  bool scaleMesh = false;
  Initfile::get("mesh->scale mesh",scaleMesh);
  if (scaleMesh) {
    Initfile::get("mesh->scale",dimension);
  } else
    dimension.set(1.0);
  
  oldSolution.resize(dow);
  for (size_t i = 0; i < dow; i++)
    oldSolution[i] = NULL;
}


NavierStokes_TaylorHood::~NavierStokes_TaylorHood() 
{ FUNCNAME("NavierStokes_TaylorHood::~NavierStokes_TaylorHood()");
  
  if (velocity != NULL) {
    delete velocity;
    velocity = NULL;
  }
  
  for (size_t i = 0; i < dow; i++) {
    if (oldSolution[i] != NULL)
      delete oldSolution[i];
    oldSolution[i] = NULL;
  }
  
  delete fileWriter;
}


void NavierStokes_TaylorHood::initData()
{ FUNCNAME("NavierStokes_TaylorHood::initTimeInterface()");

  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
  
  for (size_t i = 0; i < dow; i++)
    oldSolution[i] = new DOFVector<double>(getFeSpace(i), "old(v_"+Helpers::toString(i)+")");
  
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace()->getMesh(), velocity);

  super::initData();
}


void NavierStokes_TaylorHood::solveInitialProblem(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokes_TaylorHood::solveInitialProblem()");


  double c = 1.0;
  int initialVelocity = 0, flowDirection = 1;
  Initfile::get(name + "->initial velocity", initialVelocity);
  Initfile::get(name + "->initial velocity->value", c);
  Initfile::get(name + "->initial velocity->flow direction", flowDirection);

  std::vector<AbstractFunction<double, WorldVector<double> >*> initialFcts(2);
  if (initialVelocity == 0) { // no initial velocity
    initialFcts[0]= new ConstantFct(0.0);
    initialFcts[1]= new ConstantFct(0.0);
  } else if(initialVelocity == 1) { // constant inflow
      initialFcts[flowDirection] = new ConstantFct(c);
      initialFcts[1-flowDirection] = new ConstantFct(0.0);
  } else if(initialVelocity == 2) { // parabolic inflow
    double dist = -1.0;
    Initfile::get(name + "->initial velocity->parabolic inflow->boundary dist", dist);
    if (dist <= 0.0)
      dist = 2.0*dimension[1-flowDirection];
    initialFcts[flowDirection] = new ParabolicInflow(dist/2.0, c, flowDirection);
    initialFcts[1-flowDirection] = new ConstantFct(0.0);
  } else
    throw(std::runtime_error("Unknown initial velocity."));
  
  MSG("solve initial problem...\n");
  for (unsigned i = 0; i < dow; ++i) {
    prob->getSolution()->getDOFVector(i)->interpol(initialFcts[i]);
    prob->setExactSolutionFct(initialFcts[i], i);
  }
}


void NavierStokes_TaylorHood::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TaylorHood::transferInitialSolution()");

  for (int i = 0; i < dow; i++)
    prob->setExactSolution(prob->getSolution()->getDOFVector(i), i);

  calcVelocity();
  for (size_t i = 0; i < dow; i++)
    oldSolution[i]->copy(*prob->getSolution()->getDOFVector(i));
  
  super::transferInitialSolution(adaptInfo);
}


void NavierStokes_TaylorHood::fillOperators()
{ FUNCNAME("NavierStokes_TaylorHood::fillOperators()");

  WorldVector<DOFVector<double>* > vel;
  for (unsigned k = 0; k < dow; k++)
    vel[k] = prob->getSolution()->getDOFVector(k);
  
  // fill operators for prob
  for (unsigned i = 0; i < dow; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(getFeSpace(i), getFeSpace(i));
    opTime->addTerm(new Simple_ZOT(density));
    prob->addMatrixOperator(*opTime, i, i, getInvTau(), getInvTau());
    
    /// < (1/tau)*u_i^old , psi >
    Operator *opTimeOld = new Operator(getFeSpace(i), getFeSpace(i));
    opTimeOld->addTerm(new Phase_ZOT(getOldSolution(i), density));
    prob->addVectorOperator(*opTimeOld, i, getInvTau(), getInvTau());
 

    /// < u^old*grad(u_i^old) , psi >
    Operator *opUGradU0 = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradU0->addTerm(new WorldVec_FOT(vel, density), GRD_PHI);
    opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addVectorOperator(*opUGradU0, i);
    
    /// < u*grad(u_i^old) , psi >
    for (unsigned k = 0; k < dow; ++k) {
      Operator *opUGradU1 = new Operator(getFeSpace(i), getFeSpace(k));
      opUGradU1->addTerm(new PartialDerivative_ZOT(
	prob->getSolution()->getDOFVector(i), k, density));
      prob->addMatrixOperator(*opUGradU1, i, k);
    }

    /// < u^old*grad(u_i) , psi >
    for(unsigned k = 0; k < dow; ++k) {
      Operator *opUGradU2 = new Operator(getFeSpace(i),getFeSpace(i));
      opUGradU2->addTerm(new VecAndPartialDerivative_FOT(
	vel[k], k, density), GRD_PHI);
      prob->addMatrixOperator(*opUGradU2, i, i);
    }

    /// Diffusion-Operator
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i), getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, i, dow);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow), getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);
  }

  Operator *opNull = new Operator(getFeSpace(dow), getFeSpace(dow));
  opNull->addTerm(new Simple_ZOT(0.0));
  prob->addMatrixOperator(*opNull, dow, dow);
}


void NavierStokes_TaylorHood::addLaplaceTerm(int i)
{ FUNCNAME("NavierStokes_TaylorHood::addLaplaceTerm()");

  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (unsigned j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new MatrixIJ_SOT(j, i, viscosity));
      prob->addMatrixOperator(*opLaplaceUi1, i, j, &theta, &theta);

      if (abs(minusTheta1) > DBL_TOL) {
	opLaplaceUi1->setUhOld(prob->getSolution()->getDOFVector(j));
	prob->addVectorOperator(*opLaplaceUi1, i, &minusTheta1, &minusTheta1);
      }
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
    opLaplaceUi->addTerm(new Simple_SOT(viscosity));
  prob->addMatrixOperator(*opLaplaceUi, i, i, &theta, &theta);

  if (abs(minusTheta1) > DBL_TOL) {
    opLaplaceUi->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addVectorOperator(*opLaplaceUi, i, &minusTheta1, &minusTheta1);
  }
}


void NavierStokes_TaylorHood::fillBoundaryConditions()
{ FUNCNAME("NavierStokes_TaylorHood::fillBoundaryConditions()");

  ERROR("You have to implement some boundary conditions!\n");
}


void NavierStokes_TaylorHood::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TaylorHood::closeTimestep()");

  calcVelocity();
  for (size_t i = 0; i < dow; i++)
    oldSolution[i]->copy(*prob->getSolution()->getDOFVector(i));
  
  super::closeTimestep(adaptInfo);
}


void NavierStokes_TaylorHood::writeFiles(AdaptInfo *adaptInfo, bool force)
{ FUNCNAME("NavierStokesPhase_TaylorHood::closeTimestep()");

  super::writeFiles(adaptInfo, force);
  fileWriter->writeFiles(adaptInfo, false);
}
