#include "PhaseFieldCrystal.h"

using namespace std;
using namespace AMDiS;

PhaseFieldCrystal::PhaseFieldCrystal(const std::string &name_) :
  ProblemInstatBase(name_,NULL),
  prob(NULL), 
  useMobility(false),
  oldMeshChangeIdx(0),
  nTimesteps(-1),
  oldTimestep(0.0),
  tempParameter(-0.6),
  r(-0.4),
  rho0(1.0),
  density(-0.1),
  two(2.0),
  minus2(-2.0)
{
  // create basic problems
  prob = new ProblemStat(name + "->space");
  
  dow = Global::getGeo(WORLD);
  
  Parameters::get(name + "->r",r);
  Parameters::get(name + "->rho0", rho0);
  Parameters::get(name + "->density", density);
  Parameters::get(name + "->use mobility", useMobility);
  tempParameter= -(1.0+r);
};


PhaseFieldCrystal::~PhaseFieldCrystal() 
{ FUNCNAME("PhaseFieldCrystal::~PhaseFieldCrystal()");
};


void PhaseFieldCrystal::initialize(Flag initFlag,
                              ProblemStat *adoptProblem,
                              Flag adoptFlag)
{ FUNCNAME("PhaseFieldCrystal::initialize()");

  prob->initialize(initFlag, adoptProblem, adoptFlag);
  dim = getMesh()->getDim();
};


void PhaseFieldCrystal::initTimeInterface()
{
  fillOperators();
  fillBoundaryConditions();
};


Flag PhaseFieldCrystal::initDataFromFile(AdaptInfo *adaptInfo)
{ FUNCNAME("PhaseFieldCrystal::initDataFromFile()");

  Flag initFlag;
  int readDataFromFile = 0, readArhFiles = 0; 
  Initfile::get("user parameter->read data from file", readDataFromFile);
  Initfile::get("user parameter->read arh files", readArhFiles);
  if (readDataFromFile == 0) return initFlag;
          
  std::string pfc_file = "";  
  Initfile::get("user parameter->data file[pfc]", pfc_file);
  // read data and mesh from arh-files
  
  if (pfc_file.length() > 0) { MSG_DBG("pfc...\n");
    ArhReader::read(pfc_file, prob->getMesh(), 
      prob->getSolution()->getDOFVector(0), 
      prob->getSolution()->getDOFVector(1), 
      prob->getSolution()->getDOFVector(2));
    initFlag.setFlag(PFC_ADOPTED);
    initFlag.setFlag(PFC_MESH_ADOPTED); }

  return initFlag;
};


void PhaseFieldCrystal::solveInitialProblem(AdaptInfo *adaptInfo) 
{ FUNCNAME("PhaseFieldCrystal::solveInitialProblem()");
  MSG_DBG("PhaseFieldCrystal::solveInitialProblem()\n");

  Flag initFlag = initDataFromFile(adaptInfo);
};


void PhaseFieldCrystal::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("PhaseFieldCrystal::transferInitialSolution()");

  writeFiles(adaptInfo, false);
};


void PhaseFieldCrystal::beginIteration(AdaptInfo *adaptInfo) 
{ FUNCNAME("PhaseFieldCrystal::beginIteration()");
	
  MSG("\n");
  MSG("[[ PFC iteration ]]\n");
    
  // assemble some blocks only once, if some conditions are fullfilled
  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  bool fixedMatrixCell= false;
  #else
  bool fixedMatrixCell= adaptInfo->getTimestepNumber() > 0
    && std::abs(adaptInfo->getTimestep() - oldTimestep) < DBL_TOL
    && oldMeshChangeIdx == getMesh()->getChangeIndex();
  #endif
  for(unsigned i= 0; i < fixedMatrixTimestep.size(); ++i) {
    prob->setAssembleMatrixOnlyOnce(
      fixedMatrixTimestep[i].first, 
      fixedMatrixTimestep[i].second,
      fixedMatrixCell);
  }
};


Flag PhaseFieldCrystal::oneIteration( AdaptInfo *adaptInfo, 
                                      Flag toDo) 
{ FUNCNAME("PhaseFieldCrystal::oneIteration()");

  Flag flag, markFlag;
  
  if (toDo.isSet(MARK))
    markFlag = prob->markElements(adaptInfo);
  else
    markFlag = 3;

  // refine
  if (toDo.isSet(ADAPT) && markFlag.isSet(MESH_REFINED))
    flag = prob->refineMesh(adaptInfo);  
  // coarsen
  if (toDo.isSet(ADAPT) && markFlag.isSet(MESH_COARSENED))
    flag |= prob->coarsenMesh(adaptInfo);

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  MPI::COMM_WORLD.Barrier();
  #endif

  #ifndef NONLIN_PROBLEM
  if (toDo.isSet(BUILD))
    prob->buildAfterCoarsen(adaptInfo, markFlag, true, true);
  #endif

  if (toDo.isSet(SOLVE)) {
    try {
      MSG("solve prob..\n");
      prob->solve(adaptInfo, false);
    } catch(std::exception &e) {
      WARNING(("Could not solve system. Simulation will be stoped here! UMFPACK-ERROR: " 
        + std::string(e.what()) + "\n").c_str());
      adaptInfo->setTime(adaptInfo->getEndTime());
      adaptInfo->setTimestepNumber(adaptInfo->getNumberOfTimesteps());
      return flag;
    } catch(...) {
      WARNING("Could not solve system. Simulation will be stoped here! Unknown ERROR\n");
      adaptInfo->setTime(adaptInfo->getEndTime());
      adaptInfo->setTimestepNumber(adaptInfo->getNumberOfTimesteps());
      return flag;
    }
  }

  if (toDo.isSet(ESTIMATE))
    prob->estimate(adaptInfo);

  oldTimestep = adaptInfo->getTimestep();
  oldMeshChangeIdx = prob->getMesh()->getChangeIndex();

  return flag;
};


void PhaseFieldCrystal::endIteration(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokes_Chorin::endIteration()");
	
  MSG("\n");
  MSG("[[ end of PFC iteration ]]\n");
};


void PhaseFieldCrystal::fillOperators()
{ FUNCNAME("PhaseFieldCrystal::fillOperators()");
  MSG_DBG("PhaseFieldCrystal::fillOperators()\n");
  
  // phi*rho
  Operator *opMnew = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  opMnew->addTerm(new Simple_ZOT);
  Operator *opMold = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  opMold->addTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(0), NULL));

  // -nabla*(phi*grad(rho))
  Operator *opL = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  opL->addTerm(new Simple_SOT);
  
  Operator *opLM = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  if (useMobility) // non-constant mobility
    opLM->addTerm(new VecAtQP_SOT(
      prob->getSolution()->getDOFVector(0),
      new MobilityPfc(density, rho0)));
  else // constant mobility
    opLM->addTerm(new Simple_SOT);
  
  // -(1+r)*phi*rho
  Operator *opMTemp = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  opMTemp->addZeroOrderTerm(new Simple_ZOT());
  // -2*rho_old^3
  Operator *opMPowExpl = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  opMPowExpl->addZeroOrderTerm(new Pow3_ZOT(prob->getSolution()->getDOFVector(0),-2.0));
  // -3*rho_old^2
  Operator *opMPowImpl = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  opMPowImpl->addZeroOrderTerm(new Pow2_ZOT(prob->getSolution()->getDOFVector(0),-3.0));
  
  // 0 = nu-laplace(rho)
  // -------------------
  prob->addMatrixOperator(opL, 0, 0); // -laplace(rho)
  prob->addMatrixOperator(opMnew, 0, 2); // nu
  
  // dt(rho) = laplace(mu) - u*grad(rho)
  // -----------------------------------
  prob->addMatrixOperator(opMnew, 1, 0, getInvTau());
  prob->addMatrixOperator(opLM, 1, 1); // -laplace(mu)
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(opMold, 1, getInvTau());
  
  // mu-2*nu-laplace(nu)-(1+r)*rho = (rho_old^3) + ExtPot - eps^2/(rho_old+0.9)
  // ----------------------------------------------------------------------
  prob->addMatrixOperator(opMTemp, 2, 0, getTempParameter()); // -phi*(1+r)*rho
  prob->addMatrixOperator(opMPowImpl, 2, 0); // -3*rho*rho_old^2
  prob->addMatrixOperator(opL, 2, 0, &two); // -2*phi*laplace(rho) * psi
  prob->addMatrixOperator(opMnew, 2, 1); // phi*mu * psi
  prob->addMatrixOperator(opL, 2, 2); // phi*grad(nu) * grad(psi)
  // prob.addMatrixOperator(opMnew,2,2,&minus2);
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(opMPowExpl, 2); // -2*phi^old*rho_old^3
};


void PhaseFieldCrystal::fillBoundaryConditions()
{ FUNCNAME("PhaseFieldCrystal::fillBoundaryConditions()");
  MSG_DBG("PhaseFieldCrystal::fillBoundaryConditions()\n");
};


void PhaseFieldCrystal::initTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("PhaseFieldCrystal::initTimestep()");
};


void PhaseFieldCrystal::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("PhaseFieldCrystal::closeTimestep()");

  int outputPeriod = 1;
  Initfile::get("user parameter->write every ith timestep", outputPeriod);
  if (adaptInfo->getTimestepNumber() % outputPeriod == 0
    && adaptInfo->getStartTime() < adaptInfo->getEndTime()) {
    writeFiles(adaptInfo, false);
  }
};


void PhaseFieldCrystal::writeFiles(AdaptInfo *adaptInfo, bool force)
{
  prob->writeFiles(adaptInfo, force);
};
